# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.core.typing.callable_signature."""

import copy
import inspect
import unittest

from pyglove.core.typing import annotation_conversion   # pylint: disable=unused-import
from pyglove.core.typing import callable_signature
from pyglove.core.typing import class_schema
from pyglove.core.typing import key_specs as ks
from pyglove.core.typing import value_specs as vs


class SignatureTest(unittest.TestCase):
  """Tests for `Signature` class."""

  def test_basics(self):
    """Test basics of `Signature` class."""

    def foo(a, b: int = 1):
      del a, b

    signature = callable_signature.get_signature(foo)
    self.assertEqual(signature.module_name, 'pyglove.core.typing.callable_signature_test')
    self.assertEqual(signature.name, 'foo')
    self.assertEqual(signature.id,
                     'pyglove.core.typing.callable_signature_test.SignatureTest.test_basics.<locals>.foo')
    self.assertEqual(
        str(signature),
        'Signature(\'pyglove.core.typing.callable_signature_test.SignatureTest.test_basics.<locals>.foo\', '
        'args=[\n'
        '  Argument(name=\'a\', value_spec=Any()),\n'
        '  Argument(name=\'b\', value_spec=Any('
        'default=1, annotation=<class \'int\'>))\n])')

    self.assertEqual(signature.named_args, [
        callable_signature.Argument('a', vs.Any()),
        callable_signature.Argument(
            'b', vs.Any(default=1).annotate(int)),
    ])
    self.assertEqual(signature.arg_names, ['a', 'b'])

    # Test __eq__ and __ne__.
    def assert_not_equal(signature, field_name, modified_value):
      other = copy.copy(signature)
      setattr(other, field_name, modified_value)
      self.assertNotEqual(signature, other)

    assert_not_equal(signature, 'name', 'bar')
    assert_not_equal(signature, 'module_name', 'other_module')
    assert_not_equal(signature, 'args',
                     [signature.args[0],
                      callable_signature.Argument('b', vs.Int())])
    assert_not_equal(
        signature, 'kwonlyargs',
        list(signature.kwonlyargs) + [
            callable_signature.Argument('x', vs.Any())])
    assert_not_equal(signature, 'varargs',
                     callable_signature.Argument('args', vs.Any()))
    assert_not_equal(signature, 'varkw',
                     callable_signature.Argument('kwargs', vs.Any()))
    self.assertNotEqual(signature, 1)
    self.assertEqual(signature, signature)
    self.assertEqual(signature, copy.deepcopy(signature))

    with self.assertRaisesRegex(TypeError, '.* is not callable'):
      callable_signature.get_signature(1)

  def test_function(self):
    """Tests `get_signature` on regular functions."""

    def foo(a, b: int = 1, **kwargs):
      del a, b, kwargs

    signature = callable_signature.get_signature(foo)
    self.assertEqual(
        signature.callable_type, callable_signature.CallableType.FUNCTION)
    self.assertEqual(signature.args, [
        callable_signature.Argument('a', vs.Any()),
        callable_signature.Argument(
            'b', vs.Any(default=1).annotate(int)),
    ])
    self.assertEqual(signature.kwonlyargs, [])
    self.assertIsNone(signature.varargs)
    self.assertEqual(signature.varkw,
                     callable_signature.Argument('kwargs', vs.Any()))
    self.assertFalse(signature.has_varargs)
    self.assertTrue(signature.has_varkw)
    self.assertTrue(signature.has_wildcard_args)
    self.assertEqual(
        signature.get_value_spec('b'),
        vs.Any(default=1, annotation=int))
    # NOTE: 'x' matches **kwargs
    self.assertEqual(signature.get_value_spec('x'), vs.Any())

  def test_lambda(self):
    """Tests `get_signature` on lambda function."""
    signature = callable_signature.get_signature(lambda x: x)
    self.assertEqual(
        signature.callable_type, callable_signature.CallableType.FUNCTION)
    self.assertEqual(
        signature.args, [callable_signature.Argument('x', vs.Any())])
    self.assertEqual(signature.kwonlyargs, [])
    self.assertIsNone(signature.varargs)
    self.assertIsNone(signature.varkw)
    self.assertFalse(signature.has_varargs)
    self.assertFalse(signature.has_varkw)
    self.assertFalse(signature.has_wildcard_args)
    self.assertIsNone(signature.get_value_spec('y'))

  def test_method(self):
    """Tests get_signature on class methods."""

    class A:

      @classmethod
      def foo(cls, x: int = 1):
        return x

      def bar(self, y: int, *args, z=1):
        del args
        return y + z

      def __call__(self, z: int, **kwargs):
        del kwargs
        return z

    # Test class static method.
    signature = callable_signature.get_signature(A.foo)
    self.assertEqual(
        signature.callable_type, callable_signature.CallableType.METHOD)
    self.assertEqual(
        signature.args,
        [callable_signature.Argument(
            'x', vs.Any(default=1).annotate(int))])
    self.assertEqual(signature.kwonlyargs, [])

    # Test instance method.
    signature = callable_signature.get_signature(A().bar)
    self.assertEqual(
        signature.callable_type, callable_signature.CallableType.METHOD)
    self.assertEqual(
        signature.args,
        [callable_signature.Argument('y', vs.Any().annotate(int))])
    self.assertEqual(
        signature.kwonlyargs,
        [callable_signature.Argument('z', vs.Any(default=1))])
    self.assertEqual(
        signature.varargs,
        callable_signature.Argument('args', vs.Any()))
    self.assertTrue(signature.has_varargs)
    self.assertFalse(signature.has_varkw)

    # Test unbound instance method
    signature = callable_signature.get_signature(A.bar)
    self.assertEqual(
        signature.callable_type, callable_signature.CallableType.FUNCTION)
    self.assertEqual(signature.args, [
        callable_signature.Argument('self', vs.Any()),
        callable_signature.Argument('y', vs.Any().annotate(int))
    ])
    self.assertEqual(
        signature.kwonlyargs,
        [callable_signature.Argument('z', vs.Any(default=1))])
    self.assertEqual(
        signature.varargs,
        callable_signature.Argument('args', vs.Any()))
    self.assertTrue(signature.has_varargs)
    self.assertFalse(signature.has_varkw)

    # Test object as callable.
    signature = callable_signature.get_signature(A())
    self.assertEqual(
        signature.callable_type, callable_signature.CallableType.METHOD)
    self.assertEqual(
        signature.args,
        [callable_signature.Argument('z', vs.Any().annotate(int))])
    self.assertEqual(signature.kwonlyargs, [])
    self.assertFalse(signature.has_varargs)
    self.assertTrue(signature.has_varkw)
    self.assertEqual(
        signature.varkw,
        callable_signature.Argument('kwargs', vs.Any()))

  def test_make_function(self):
    """Tests `Signature.make_function`."""

    def func1(x, y=1) -> int:
      del x, y

    def func2(x=1, *, y):
      del x, y

    def func3(x=1, *y):    # pylint: disable=keyword-arg-before-vararg
      del x, y

    def func4(*y):
      del y

    def func5(*, x=1, y):
      del x, y

    def func6(x=1, *, y, **z) -> str:
      del x, y, z

    for func in [func1, func2, func3, func4, func5, func6]:
      new_func = callable_signature.get_signature(func).make_function(['pass'])
      old_signature = inspect.signature(func)
      new_signature = inspect.signature(new_func)
      self.assertEqual(old_signature, new_signature)


class FromSchemaTest(unittest.TestCase):
  """Tests for `Signature.from_schema`."""

  def _get_signature(self, init_arg_list, is_method: bool = True):
    s = class_schema.Schema([
        class_schema.Field('x', vs.Int(), 'x'),
        class_schema.Field('y', vs.Int(), 'y'),
        # Frozen fields will be ignored.
        class_schema.Field('v', vs.Bool().freeze(True), 'v'),
        class_schema.Field('z', vs.List(vs.Int()), 'z'),
        class_schema.Field(ks.StrKey(), vs.Str(), 'kwargs'),
    ], metadata=dict(init_arg_list=init_arg_list), allow_nonconst_keys=True)
    return callable_signature.Signature.from_schema(
        s, 'bar', 'foo', is_method=is_method)

  def test_classmethod_with_regular_args(self):
    self.assertEqual(
        self._get_signature(['x', 'y', 'z']),
        callable_signature.Signature(
            callable_type=callable_signature.CallableType.FUNCTION,
            module_name='bar',
            name='foo',
            args=[
                callable_signature.Argument('self', vs.Any()),
                callable_signature.Argument('x', vs.Int()),
                callable_signature.Argument('y', vs.Int()),
                callable_signature.Argument('z', vs.List(vs.Int())),
            ],
            varkw=callable_signature.Argument('kwargs', vs.Str())))

  def test_function_with_varargs(self):
    self.assertEqual(
        self._get_signature(['x', '*z'], is_method=False),
        callable_signature.Signature(
            callable_type=callable_signature.CallableType.FUNCTION,
            module_name='bar',
            name='foo',
            args=[
                callable_signature.Argument('x', vs.Int()),
            ],
            kwonlyargs=[
                callable_signature.Argument('y', vs.Int()),
            ],
            varargs=callable_signature.Argument('z', vs.Int()),
            varkw=callable_signature.Argument('kwargs', vs.Str())))

  def test_classmethod_with_kwonly_args(self):
    self.assertEqual(
        self._get_signature([]),
        callable_signature.Signature(
            callable_type=callable_signature.CallableType.FUNCTION,
            module_name='bar',
            name='foo',
            args=[
                callable_signature.Argument('self', vs.Any()),
            ],
            kwonlyargs=[
                callable_signature.Argument('x', vs.Int()),
                callable_signature.Argument('y', vs.Int()),
                callable_signature.Argument(
                    'z', vs.List(vs.Int())),
            ],
            varkw=callable_signature.Argument('kwargs', vs.Str())))

  def test_bad_cases(self):
    with self.assertRaisesRegex(
        ValueError,
        'Variable positional argument \'x\' should have a value of '
        '`pg.typing.List` type'):
      _ = self._get_signature(['*x'])

    with self.assertRaisesRegex(
        ValueError,
        'Argument \'a\' is not a symbolic field.'):
      _ = callable_signature.Signature.from_schema(
          class_schema.Schema([], metadata=dict(init_arg_list=['a'])),
          '__main__', 'foo')

    class Foo:
      __call__ = 1

    with self.assertRaisesRegex(
        TypeError, '.*__call__ is not a method'):
      callable_signature.Signature.from_callable(Foo)


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.hyper.Evolvable."""

import random
import re
import unittest

from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core.hyper.evolvable import evolve
from pyglove.core.hyper.evolvable import MutationType


class Layer(symbolic.Object):
  pass


@symbolic.members([
    ('layers', pg_typing.List(pg_typing.Object(Layer))),
])
class Sequential(Layer):
  pass


class Activation(Layer):
  pass


class ReLU(Activation):
  pass


class Swish(Activation):
  pass


@symbolic.members([
    ('filters', pg_typing.Int(min_value=1)),
    # `kernel_size` is marked as no_mutation, which should not appear as a
    # mutation candidate.
    ('kernel_size', pg_typing.Int(min_value=1), '', {'no_mutation': True}),
    ('activation', pg_typing.Object(Activation).noneable())
])
class Conv(Layer):
  pass


def seed_program():
  return Sequential([
      Conv(16, 3, ReLU()),
      Conv(32, 5, Swish()),
      Sequential([
          Conv(64, 7)
      ])
  ])


def mutate_at_location(mutation_type: MutationType, location: str):
  def _weights(mt, k, v, p):
    del v, p
    if mt == mutation_type and re.match(location, str(k)):
      return 1.0
    return 0.0
  return _weights


class EvolvableTest(unittest.TestCase):
  """Tests for pg.hyper.Evolvable."""

  def test_basics(self):
    v = evolve(
        seed_program(), lambda k, v, p: ReLU(),
        weights=mutate_at_location(MutationType.REPLACE, r'^layers\[.*\]$'))
    self.assertEqual(
        seed_program(), v.custom_decode(v.custom_encode(seed_program())))
    self.assertEqual(v.first_dna(), v.custom_encode(seed_program()))
    self.assertEqual(v.random_dna(), v.custom_encode(seed_program()))
    self.assertEqual(
        v.random_dna(random.Random(1), v.first_dna()),
        v.custom_encode(
            Sequential([
                ReLU(),
                Conv(32, 5, Swish()),
                Sequential([
                    Conv(64, 7)
                ])
            ])))

  def test_replace(self):
    v = evolve(
        seed_program(), lambda k, v, p: ReLU(),
        weights=mutate_at_location(MutationType.REPLACE, r'^layers\[1\]$'))
    self.assertEqual(
        v.mutate(seed_program()),
        Sequential([
            Conv(16, 3, ReLU()),
            ReLU(),
            Sequential([
                Conv(64, 7)
            ])
        ]))

    # Mutating at root.
    v = evolve(
        seed_program(), lambda k, v, p: ReLU(),
        weights=lambda mt, k, v, p: 1.0 if p is None else 0.0)
    self.assertEqual(
        v.mutate(seed_program()),
        ReLU()
    )

  def test_insertion(self):
    v = evolve(
        seed_program(), lambda k, v, p: ReLU(),
        weights=mutate_at_location(MutationType.INSERT, r'^layers\[1\]$'))
    self.assertEqual(
        v.mutate(seed_program()),
        Sequential([
            Conv(16, 3, ReLU()),
            ReLU(),
            Conv(32, 5, Swish()),
            Sequential([
                Conv(64, 7)
            ])
        ]))

  def test_delete(self):
    v = evolve(
        seed_program(), lambda k, v, p: ReLU(),
        weights=mutate_at_location(MutationType.DELETE, r'^layers\[1\]$'))
    self.assertEqual(
        v.mutate(seed_program(), random.Random(1)),
        Sequential([
            Conv(16, 3, ReLU()),
            Sequential([
                Conv(64, 7)
            ])
        ]))

  def test_random_generator(self):
    v = evolve(
        seed_program(), lambda k, v, p: ReLU(),
        weights=mutate_at_location(MutationType.REPLACE, r'^layers\[.*\]$'))
    self.assertEqual(
        v.mutate(seed_program(), random_generator=random.Random(1)),
        Sequential([
            ReLU(),
            Conv(32, 5, Swish()),
            Sequential([
                Conv(64, 7)
            ])
        ]))

  def test_mutation_points_and_weights(self):
    v = evolve(seed_program(), lambda k, v, p: v, weights=lambda *x: 1.0)
    points, weights = v.mutation_points_and_weights(seed_program())

    # NOTE(daiyip): Conv.kernel_size is marked with 'no_mutation', thus
    # it should not show here.
    self.assertEqual([(p.mutation_type, p.location) for p in points], [
        (MutationType.REPLACE, ''),
        (MutationType.REPLACE, 'layers'),
        (MutationType.INSERT, 'layers[0]'),
        (MutationType.DELETE, 'layers[0]'),
        (MutationType.REPLACE, 'layers[0]'),
        (MutationType.REPLACE, 'layers[0].filters'),
        (MutationType.REPLACE, 'layers[0].activation'),
        (MutationType.INSERT, 'layers[1]'),
        (MutationType.DELETE, 'layers[1]'),
        (MutationType.REPLACE, 'layers[1]'),
        (MutationType.REPLACE, 'layers[1].filters'),
        (MutationType.REPLACE, 'layers[1].activation'),
        (MutationType.INSERT, 'layers[2]'),
        (MutationType.DELETE, 'layers[2]'),
        (MutationType.REPLACE, 'layers[2]'),
        (MutationType.REPLACE, 'layers[2].layers'),
        (MutationType.INSERT, 'layers[2].layers[0]'),
        (MutationType.DELETE, 'layers[2].layers[0]'),
        (MutationType.REPLACE, 'layers[2].layers[0]'),
        (MutationType.REPLACE, 'layers[2].layers[0].filters'),
        (MutationType.REPLACE, 'layers[2].layers[0].activation'),
        (MutationType.INSERT, 'layers[2].layers[1]'),
        (MutationType.INSERT, 'layers[3]'),
    ])
    self.assertEqual(weights, [1.0] * len(points))

  def test_mutation_points_and_weights_with_honoring_list_size(self):
    # Non-typed list. There is no size limit.
    v = evolve(
        symbolic.List([]), lambda k, v, p: v,
        weights=lambda *x: 1.0)
    points, _ = v.mutation_points_and_weights(symbolic.List([1]))
    self.assertEqual([(p.mutation_type, p.location) for p in points], [
        (MutationType.REPLACE, ''),
        (MutationType.INSERT, '[0]'),
        (MutationType.DELETE, '[0]'),
        (MutationType.REPLACE, '[0]'),
        (MutationType.INSERT, '[1]'),
    ])

    # Typed list with size limit.
    value_spec = pg_typing.List(pg_typing.Int(), min_size=1, max_size=3)
    points, _ = v.mutation_points_and_weights(
        symbolic.List([1, 2], value_spec=value_spec))
    self.assertEqual([(p.mutation_type, p.location) for p in points], [
        (MutationType.REPLACE, ''),
        (MutationType.INSERT, '[0]'),
        (MutationType.DELETE, '[0]'),
        (MutationType.REPLACE, '[0]'),
        (MutationType.INSERT, '[1]'),
        (MutationType.DELETE, '[1]'),
        (MutationType.REPLACE, '[1]'),
        (MutationType.INSERT, '[2]'),
    ])
    points, _ = v.mutation_points_and_weights(
        symbolic.List([1], value_spec=value_spec))
    self.assertEqual([(p.mutation_type, p.location) for p in points], [
        (MutationType.REPLACE, ''),
        (MutationType.INSERT, '[0]'),
        (MutationType.REPLACE, '[0]'),
        (MutationType.INSERT, '[1]'),
    ])
    points, _ = v.mutation_points_and_weights(
        symbolic.List([1, 2, 3], value_spec=value_spec))
    self.assertEqual([(p.mutation_type, p.location) for p in points], [
        (MutationType.REPLACE, ''),
        (MutationType.DELETE, '[0]'),
        (MutationType.REPLACE, '[0]'),
        (MutationType.DELETE, '[1]'),
        (MutationType.REPLACE, '[1]'),
        (MutationType.DELETE, '[2]'),
        (MutationType.REPLACE, '[2]'),
    ])


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.object_utils.formatting."""
import inspect
import unittest

from pyglove.core.object_utils import common_traits
from pyglove.core.object_utils import formatting


class StringHelperTest(unittest.TestCase):
  """Tests for string helper methods in formatting."""

  def test_kvlist_str(self):
    self.assertEqual(
        formatting.kvlist_str([
            ('', 'foo', None),
            ('a', 1, None),
            ('b', 'str', (None, 'str')),
            ('c', True, False),
        ]), 'foo, a=1, c=True')

  def test_quote_if_str(self):
    self.assertEqual(formatting.quote_if_str(1), 1)
    self.assertEqual(formatting.quote_if_str('foo'), '\'foo\'')
    self.assertEqual(formatting.quote_if_str('foo\'s\na'), '"foo\'s\\na"')

  def test_message_on_path(self):
    self.assertEqual(formatting.message_on_path('hi.', None), 'hi.')
    self.assertEqual(
        formatting.message_on_path('hi.', formatting.KeyPath()),
        'hi. (path=)')

  def test_comma_delimited_str(self):
    self.assertEqual(
        formatting.comma_delimited_str([1, 2, 'abc']), '1, 2, \'abc\'')

  def test_auto_plural(self):
    self.assertEqual(formatting.auto_plural(2, 'number'), 'numbers')
    self.assertEqual(formatting.auto_plural(2, 'was', 'were'), 'were')


class FormatTest(unittest.TestCase):
  """Tests for formatting.format."""

  def test_formattable(self):

    class A(common_traits.Formattable):

      def format(self, compact=True, **kwargs):
        if compact:
          return 'A()'
        else:
          return 'A(...)'

    self.assertEqual(str(A()), 'A(...)')
    self.assertEqual(repr(A()), 'A()')

  def test_simple_types(self):
    self.assertEqual(formatting.format(True, compact=True), 'True')
    self.assertEqual(formatting.format(1, compact=True), '1')
    self.assertEqual(formatting.format(1.0, compact=True), '1.0')
    self.assertEqual(formatting.format('foo', compact=True), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', compact=True), '"foo\'s\\na"')

    # Compact=False has no impact on simple types.
    self.assertEqual(formatting.format(True, compact=False), 'True')
    self.assertEqual(formatting.format(1, compact=False), '1')
    self.assertEqual(formatting.format(1.0, compact=False), '1.0')
    self.assertEqual(formatting.format('foo', compact=False), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', compact=False), '"foo\'s\\na"')

    # Verbose has no impact on simple types.
    self.assertEqual(formatting.format(True, verbose=True), 'True')
    self.assertEqual(formatting.format(1, verbose=True), '1')
    self.assertEqual(formatting.format(1.0, verbose=True), '1.0')
    self.assertEqual(formatting.format('foo', verbose=True), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', verbose=True), '"foo\'s\\na"')

    # Root indent has no impact on simple types.
    self.assertEqual(formatting.format(True, root_indent=4), 'True')
    self.assertEqual(formatting.format(1, root_indent=4), '1')
    self.assertEqual(formatting.format(1.0, root_indent=4), '1.0')
    self.assertEqual(formatting.format('foo', root_indent=4), '\'foo\'')
    self.assertEqual(
        formatting.format('foo\'s\na', root_indent=4), '"foo\'s\\na"')

  def test_complex_types(self):

    class CustomFormattable(common_traits.Formattable):
      """Custom formattable."""

      def format(self, custom_param=None, **kwargs):
        return f'CustomFormattable({custom_param})'

    class A:
      pass

    self.assertEqual(
        formatting.format(
            {
                'a': CustomFormattable(),
                'b': {
                    'c': [1, 2, 3],
                    'd': ['foo', 'bar\na', 3, 4, 5]
                }
            },
            compact=True,
            custom_param='foo'),
        "{'a': CustomFormattable(foo), 'b': {'c': [1, 2, 3], "
        "'d': ['foo', 'bar\\na', 3, 4, 5]}}")

    self.assertEqual(
        formatting.format(
            {
                'a': A(),
                'b': {
                    'c': [1, 2, 3],
                    'd': ['foo', 'bar\na', 3, 4, 5]
                }
            },
            compact=False,
            list_wrap_threshold=15,
            strip_object_id=True),
        inspect.cleandoc("""{
          'a': A(...),
          'b': {
            'c': [1, 2, 3],
            'd': [
              'foo',
              'bar\\na',
              3,
              4,
              5
            ]
          }
        }"""))

  def test_include_exclude_keys(self):
    """Test format with excluded keys."""

    class A:
      pass

    class B(common_traits.Formattable):
      """Custom formattable."""

      def format(
          self, custom_param=None,
          include_keys=None, exclude_keys=None, **kwargs):
        exclude_keys = exclude_keys or set()
        kv = dict(a=1, b=2, c=3)
        def _should_include(k):
          if include_keys:
            return k in include_keys
          return k not in exclude_keys
        kv_pairs = [(k, v, None) for k, v in kv.items() if _should_include(k)]
        return f'B({formatting.kvlist_str(kv_pairs, compact=True)})'

    self.assertEqual(
        formatting.format(B(), compact=False, include_keys=set(['a', 'c'])),
        'B(a=1, c=3)')
    self.assertEqual(
        formatting.format(B(), compact=False, exclude_keys=set(['a', 'c'])),
        'B(b=2)')
    self.assertEqual(
        formatting.format(
            {
                'a': A(),
                'b': B(),
                'c': {
                    'd': [1, 2, 3],
                }
            },
            compact=False,
            list_wrap_threshold=15,
            strip_object_id=True,
            # 'a' should be removed, but 'b.a', 'c.d' should be kept as they are
            # not at the top level.
            exclude_keys=set(['a', 'd'])),
        inspect.cleandoc("""{
          'b': B(a=1, b=2, c=3),
          'c': {
            'd': [1, 2, 3]
          }
        }"""))

  def test_markdown(self):
    self.assertEqual(
        formatting.format([1], compact=True, markdown=True), '`[1]`'
    )
    self.assertEqual(
        formatting.format(
            [1, 2, 3], list_wrap_threshold=5, compact=False, markdown=True
        ),
        inspect.cleandoc("""
            ```
            [
              1,
              2,
              3
            ]
            ```
            """),
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.symbolize."""

import unittest

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import list as pg_list  # pylint: disable=unused-import
from pyglove.core.symbolic import schema_utils


class CallableSchemaTest(unittest.TestCase):
  """Tests for `callable_schema`."""

  def test_function_schema(self):
    def foo(x: int, *args, y: str, **kwargs) -> float:
      """A function.

      Args:
        x: Input 1.
        *args: Variable positional args.
        y: Input 2.
        **kwargs: Variable keyword args.

      Returns:
        The result.
      """
      del x, y, args, kwargs

    schema = schema_utils.callable_schema(foo, auto_typing=True, auto_doc=True)
    self.assertEqual(schema.name, f'{foo.__module__}.{foo.__qualname__}')
    self.assertEqual(schema.description, 'A function.')
    self.assertEqual(
        list(schema.fields.values()),
        [
            pg_typing.Field('x', pg_typing.Int(), description='Input 1.'),
            pg_typing.Field(
                'args',
                pg_typing.List(pg_typing.Any(), default=[]),
                description='Variable positional args.',
            ),
            pg_typing.Field('y', pg_typing.Str(), description='Input 2.'),
            pg_typing.Field(
                pg_typing.StrKey(),
                pg_typing.Any(),
                description='Variable keyword args.',
            ),
        ],
    )

  def test_class_init_schema(self):
    class A:

      def __init__(self, x: int, *args, y: str, **kwargs) -> float:
        """Constructor.

        Args:
          x: Input 1.
          *args: Variable positional args.
          y: Input 2.
          **kwargs: Variable keyword args.

        Returns:
          The result.
        """
        del x, y, args, kwargs

    schema = schema_utils.callable_schema(
        A.__init__, auto_typing=True, auto_doc=True, remove_self=True
    )
    self.assertEqual(schema.name, f'{A.__module__}.{A.__init__.__qualname__}')
    self.assertEqual(schema.description, 'Constructor.')
    self.assertEqual(
        list(schema.fields.values()),
        [
            pg_typing.Field('x', pg_typing.Int(), description='Input 1.'),
            pg_typing.Field(
                'args',
                pg_typing.List(pg_typing.Any(), default=[]),
                description='Variable positional args.',
            ),
            pg_typing.Field('y', pg_typing.Str(), description='Input 2.'),
            pg_typing.Field(
                pg_typing.StrKey(),
                pg_typing.Any(),
                description='Variable keyword args.',
            ),
        ],
    )


class SchemaDescriptionFromDocStrTest(unittest.TestCase):
  """Tests for `schema_description_from_docstr`."""

  def test_none_doc_str(self):
    self.assertIsNone(schema_utils.schema_description_from_docstr(None))

  def test_short_description_only(self):
    docstr = object_utils.DocStr.parse(
        """This is a function.""")
    self.assertEqual(
        schema_utils.schema_description_from_docstr(docstr),
        'This is a function.')
    self.assertEqual(
        schema_utils.schema_description_from_docstr(
            docstr, include_long_description=True),
        'This is a function.')

  def test_long_description_only(self):
    docstr = object_utils.DocStr.parse(
        """This is a function.
        
        This is the longer explanation of the function.
        
        """)
    self.assertEqual(
        schema_utils.schema_description_from_docstr(docstr),
        'This is a function.')
    self.assertEqual(
        schema_utils.schema_description_from_docstr(
            docstr, include_long_description=True),
        ('This is a function.\n\n'
         'This is the longer explanation of the function.'))


if __name__ == '__main__':
  unittest.main()
